/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file bitfield.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in April 1999. Last revision: October 17, 2005.


#ifndef _CAPD_BITFIELD_H_
#define _CAPD_BITFIELD_H_

#include "homology/config.h"

#include <iostream>
#include <cstdlib>

namespace capd {
namespace homology {


// classes defined within this header file:
class BitField;
class SetOfBitFields;

// shorter or lower-case versions of the class names
typedef BitField bitfield;
typedef SetOfBitFields bitfieldset;


// --------------------------------------------------
// ------------------- BitField ---------------------
// --------------------------------------------------

class BitField
{
public:
	// constructor of an empty BitField
	BitField ();

	// destructor
	~BitField ();

	// check if the bit field has already been defined
	int defined () const;

	// define the bit field as a piece of a larger table
	void define (unsigned char *buf, int length);

	// allocate and free memory for a bit field
	void allocate (int length);
	void free ();

	// set, clear and test bits
	void set (int n);
	void clear (int n);
	int test (int n) const;
	void takebits (const BitField &from, int length);
	void clearall (int length);

	// find the first bit set beginning at the given one
	// and return -1 if not found
	int find (int after, int length) const;

	// keys for hashing
	unsigned hashkey (int length) const;
	unsigned hashadd (int length) const;

	// comparison of bit fields
	friend int equals (const BitField &b1, const BitField &b2,
		int length);

	// conversion to/from an integer number
	friend void int2bits (int bits, int length, BitField &field);
	friend int bits2int (const BitField &field, int length);

protected:
	// the table of 8-bit cells
	unsigned char *table;

}; /* BitField */

// --------------------------------------------------

inline BitField::BitField ()
{
	table = NULL;
	return;
} /* BitField::BitField */

inline int BitField::defined () const
{
	return !!table;
} /* BitField::defined */

inline void BitField::free ()
{
	delete [] table;
	table = NULL;
	return;
} /* BitField::free */

inline BitField::~BitField ()
{
	return;
} /* BitField::~BitField */

inline void BitField::set (int n)
{
	table [n >> 3] |= (unsigned char) (0x01 << (n & 0x07));
	return;
} /* BitField::set */

inline void BitField::clear (int n)
{
	table [n >> 3] &= (unsigned char) ~(0x01 << (n & 0x07));
	return;
} /* BitField::clear */

inline int BitField::test (int n) const
{
	return !!(table [n >> 3] & (0x01 << (n & 0x07)));
} /* BitField::test */

inline void int2bits (int bits, int length, BitField &field)
{
	unsigned char *tab = field. table;
	if (!tab)
		throw "Trying to set values to an undefined bitfield.";
	while (length >= 0)
	{
		*(tab ++) = (unsigned char) (bits & 0xFF);
		bits >>= 8;
		length -= 8;
	}
	return;
} /* int2bits */

inline int bits2int (const BitField &field, int length)
{
	const unsigned char *tab = field. table;
	if (!tab)
		throw "Trying to set values to an undefined bitfield.";
	int n = 0;
	int shiftvalue = 0;
	while (length >= 8)
	{
		n |= (*(tab ++)) << shiftvalue;
		length -= 8;
		shiftvalue += 8;
	}
	const int bitmasks [] = {0x00, 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF};
	if (length > 0)
		n |= ((*(tab ++)) & bitmasks [length]) << shiftvalue;
	return n;
} /* bits2int */


// --------------------------------------------------
// ----------------- SetOfBitFields -----------------
// --------------------------------------------------

class SetOfBitFields
{
public:
	// constructor: you must submit the length of each BitField
	// and the [maximal] number of BitFields in the set
	SetOfBitFields (int len, int max);

	// destructor
	~SetOfBitFields ();

	// add a bit field; if it already was, return its value;
	// if no room available, return -1
	int add (const BitField &b, int value);

	// get a bit field's value or return -1 if not defined
	int check (const BitField &b) const;

	// how many bit fields are in use?
	int used (void) const;

	// forget all the bit fields
	void forget ();

protected:
	// number of bit fields that still can be stored
	int avail;

	// number of bit fields used
	int usedcount;

	// actual size of the table
	unsigned size;

	// length of bit fields
	int length;

	// table of bit fields
	BitField *bf;

	// BitFields' values
	BitField values;

	// buffer used for bit fields
	unsigned char *buf;

	// current position in the buffer
	unsigned char *bufcur;

}; /* class SetOfBitFields */

// --------------------------------------------------

inline SetOfBitFields::~SetOfBitFields ()
{
	if (bf)
		delete [] bf;
	if (buf)
		delete [] buf;
	if (length)
		values. free ();
	return;
} /* SetOfBitFields::~SetOfBitFields */

inline int SetOfBitFields::used () const
{
	return usedcount;
} /* SetOfBitFields::used */

inline void SetOfBitFields::forget ()
{
	if (bf)
		delete [] bf;
	bf = NULL;
	if (buf)
		delete [] buf;
	buf = NULL;
	if (length)
		values. free ();
	length = 0;
	size = 0;
	avail = 0;
	return;
} /* SetOfBitFields::forget */

} // namespace homology
} // namespace capd

#endif // _BITFIELD_H_

/// @}
