/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file bitmaps.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in June 1998. Last revision: September 3, 2003 (Nov 13, 2003).

// WARNING: This version of the module has very limited capabilities,
// e.g. it can normally write only 1-bit B/W and 24-bit true color bitmaps,
// and operates improperly if the entire output bitmap is not stored
// in the memory or is not written row by row.


#ifndef _CAPD_BITMAPS_H_
#define _CAPD_BITMAPS_H_

#include "homology/config.h"
#include "homology/textfile.h"

#include <iostream>
#include <stdio.h>
#include <cstdlib>

namespace capd {
namespace homology {


// classes defined within this header file:
class bmpfile;


// --------------------------------------------------
// -------------------- bmpfile ---------------------
// --------------------------------------------------

class bmpfile
{
public:
	// the only allowed constructor
	bmpfile (int _rows = 20);

	// default destructor
	~bmpfile ();

#ifndef ReadOnly
#define ReadOnly 0
#endif

#ifndef ReadWrite
#define ReadWrite 1
#endif
	// open a bmp file for reading (mode = ReadOnly)
	// or editing (mode = 1);
	// return: 0 = Ok, -1 = Not found, -2 = Can't read header,
	// -3 = Can't read palette, -4 = Not BMP, -5 = Not supported,
	// -6 = Memory allocation error occurred
	int open (const char *filename, int mode = ReadOnly,
		int quiet = 0);

	// create a bmp file (only 1- and 24-bit supported - sorry!);
	// set "x_resolution" and "y_resolution" before or 0 will be
	// assumed; if you change these parameters later, you must
	// rewrite the header;
	// return: 0 = Ok, -1 = Can't create,
	// -2 = Can't write header, -3 = Can't write palette,
	// -4 = Bad params, -5 = Not supported,
	// -6 = Memory allocation error occurred
	int create (const char *filename, int _width, int _height,
		int _bitsperpixel = 24, int _pallength = 0,
		int32 *_pallette = NULL, int quiet = 0);

	// should the inverted picture be processed,
	// as stored in the file?
	inline void invertedpicture (int _inverted = 1)
		{inverted = _inverted;};

	// write the header to the file;
	// return: 0 = Ok, -1 = Error (no messages displayed)
	int writeheader (void);

	// read the true color of a pixel or its index if 'index';
	// note: the coordinates start with 0
	int32 getpixel (int x, int y, bool index = false);

	// write a pixel and return its true color
	int32 putpixel (int x, int y, long color = 0,
		bool index = false);

	// write a pixel and return its true color
	void drawpixel (int x, int y, int thickness = 1,
		long color = 0, bool index = false);

	// draw a segment built of pixels, with the desired thickness
	void drawsegment (int x1, int y1, int x2, int y2,
		int thickness = 1, long color = 0,
		bool index = false);

	// get the size of the picture
	inline int picture_width (void) { return width; };
	inline int picture_height (void) { return height; };

	// the resolution of the picture
	int32 x_resolution;
	int32 y_resolution;

	// statistics
	long rows_read;
	long rows_written;

	// flushing buffers
	void flush (void);

	// get the current number of bits per pixel
	int getbitsperpixel (void);

	// get the pointer to the given row stored in the memory
	unsigned char *getrow (int row);
	
	// get the length of a row
	unsigned getrowlength () const;

	// get the pointer to the current color palette (if any)
	inline int32 *getpalette (void) { return palette; };

	// get the length of the current palette (if any)
	inline int getpallength (void) { return pallength; };

protected:
	// the header of the bitmap file
	char *header;

	// palette length (in 32-bit words) and the palette itself
	int pallength;
	int32 *palette;
	bool palette_allocated;

	// rows of the picture stored in the memory
	// and the current one
	int *rownumber;
	unsigned char **rowbuf;
	char *changed;
	unsigned char *defined;
	int current;

	// read and write the current row
	int readrow (int row);
	int writerow (int row);

	// where the bitmap starts in the file
	int bmpstart;

	// the length of a row in the bitmap (in bytes)
	unsigned rowlength;

	// number of bits per pixel in the bitmap file
	int bitsperpixel;

	// the actual size of the picture stored in the file
	int width, height;

	// the binary stream containing the picture
	FILE *fbmp;

	// the type of the file: 0 = none, 1 = read,
	// 2 = write, 3 = r/w
	int fileinuse;

	// the number of the first row undefined in the picture;
	// normally it will be the height of the picture, since
	// rows are numbered starting with 0; it is important
	// while creating a BMP picture
	int firstundefinedrow;

	// the number of the first undefined byte in the BMP file
	int firstundefinedbyte;

	// number of rows allocated in the memory
	int rows;

	// byte to fill empty lines
	unsigned char fillbyte;

	// should the inverted picture be processed
	// as stored in the file?
	int inverted;

}; /* class bmpfile */

} // namespace homology
} // namespace capd

#endif // _BITMAPS_H_

/// @}
