/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cubebase.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: May 17, 2006.


#ifndef _CAPD_CUBEBASE_H_
#define _CAPD_CUBEBASE_H_

#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/pointset.h"
#include "homology/chains.h"
#include "homology/bitfield.h"
#include "homology/integer.h"
#include "homology/hashsets.h"
#include "homology/gcomplex.h"
#include "homology/pointbas.h"
#include "homology/cubemain.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstring>

namespace capd {
namespace homology {

// a friend class template
template <class coordtype>
class tCellBase;

// --------------------------------------------------
// -------------- Cube with PointBase ---------------
// --------------------------------------------------
// An n-dimensional cube of size 1 in R^n
// with integer coordinates and edges parallel to the axes.

template <class coordtype>
class tCubeBase
{
public:
	// the type of the coordinates
	typedef coordtype CoordType;
	
	// the maximal dimension of a cube
	static const int MaxDim = MaxBasDim;

	// default constructor
	tCubeBase ();
	
	// a constructor from a table of coordinates
	tCubeBase (const coordtype *c, int dim);

	// a constructor from a number (should be used with caution)
	tCubeBase (int number, int dim);

	// copy constructor
	tCubeBase (const tCubeBase<coordtype> &c);

	// assignment operator
	tCubeBase<coordtype> &operator =
		(const tCubeBase<coordtype> &c);

	// return the dimension of the cube
	int dim () const;

	// return the dimension of the cube
	operator int () const;

	// return the coordinates of the cube
	const coordtype *coord () const;

	// fill out the coordinate table with the cube's coordinates
	template <class intType>
	intType *coord (intType *c) const;

	// return the coordinates of the cube
	operator const coordtype * () const;

	// hashing key no. 1 required by the hashing set template
	unsigned hashkey1 () const;

	// hashing key no. 2 required by the hashing set template
	unsigned hashkey2 () const;

	// the name of the objects represented by this class
	static const char *name ();

	// the plural name of the objects represented by this class
	static const char *pluralname ();

	// friends
	friend inline int operator == (const tCubeBase<coordtype> &c1,
		const tCubeBase<coordtype> &c2)
	{
		return c1. n == c2. n;
	} /* operator == */


//	template <class _coordtype>
//	friend int operator == (const tCubeBase<_coordtype> &c1,
//		const tCubeBase<_coordtype> &c2);

	template <class _coordtype>
	friend int operator != (const tCubeBase<_coordtype> &c1,
		const tCubeBase<_coordtype> &c2);
	
	friend class tCellBase<coordtype>;

private:
	// return the actual number of the cube in the point base
	int num () const;

	// high bits = space dimension, the remaining bits
	// form the number of the cube in an appropriate pointset
	int n;

}; /* class tCubeBase */

// --------------------------------------------------

template <class coordtype>
inline int tCubeBase<coordtype>::num () const
{
	return (n & NumMask);
} /* tCubeBase::num */

// --------------------------------------------------

template <class coordtype>
inline tCubeBase<coordtype>::tCubeBase ()
: n (0)
{
	return;
} /* tCubeBase::tCubeBase */

template <class coordtype>
inline tCubeBase<coordtype>::tCubeBase (const coordtype *c,
	int dim)
{
	if (dim <= 0)
		throw "Non-positive dimension of a cube.";
	if (dim >= MaxDim)
		throw "Too high space dimension.";
	n = tPointBase<coordtype>::number (c, dim);
	if (n < 0)
		throw "Negative number of a cube.";
	n |= (dim << NumBits);
	return;
} /* tCubeBase::tCubeBase */

template <class coordtype>
inline tCubeBase<coordtype>::tCubeBase (int number, int dim)
{
	n = number | (dim << NumBits);
	return;
} /* tCubeBase::tCubeBase */

template <class coordtype>
inline tCubeBase<coordtype>::tCubeBase (const tCubeBase<coordtype> &c)
{
	n = c. n;
	return;
} /* tCubeBase::tCubeBase */

template <class coordtype>
inline tCubeBase<coordtype> &tCubeBase<coordtype>::operator =
	(const tCubeBase<coordtype> &c)
{
	n = c. n;
	return *this;
} /* tCubeBase::operator = */

template <class coordtype>
inline int tCubeBase<coordtype>::dim () const
{
	return (int) (n >> NumBits);
} /* tCubeBase::dim */

template <class coordtype>
inline tCubeBase<coordtype>::operator int () const
{
	return dim ();
} /* tCubeBase::operator int */

template <class coordtype>
inline const coordtype *tCubeBase<coordtype>::coord () const
{
	return tPointBase<coordtype>::coord (num (), dim ());
} /* tCubeBase::coord */

template <class coordtype>
template <class intType>
inline intType *tCubeBase<coordtype>::coord (intType *c) const
{
	int d = dim ();
	const coordtype *tab = tPointBase<coordtype>::coord (num (), d);
	for (int i = 0; i < d; ++ i)
		c [i] = static_cast<intType> (tab [i]);
	return c;
} /* tCubeBase::coord */

template <class coordtype>
inline tCubeBase<coordtype>::operator const coordtype * () const
{
	return coord ();
} /* tCubeBase::operator coordtype * */

template <class coordtype>
inline unsigned tCubeBase<coordtype>::hashkey1 () const
{
	return ((n & 0x555555) << 17) + ((n & 0xAA00AA) << 7) +
		((n & 0xAA00) >> 7);
} /* tCubeBase::hashkey1 */

template <class coordtype>
inline unsigned tCubeBase<coordtype>::hashkey2 () const
{
	return ((n & 0xAAAAAA) << 18) + ((n & 0x55555500) >> 8) +
		((n & 0x555555) << 10);
} /* tCubeBase::hashkey2 */

template <class coordtype>
inline const char *tCubeBase<coordtype>::name ()
{
	return "cube";
} /* tCubeBase::name */

template <class coordtype>
inline const char *tCubeBase<coordtype>::pluralname ()
{
	return "cubes";
} /* tCubeBase::pluralname */

// --------------------------------------------------

/*
template <class coordtype>
inline int operator == (const tCubeBase<coordtype> &c1,
	const tCubeBase<coordtype> &c2)
{
	return c1. n == c2. n;
}*/ /* operator == */

template <class coordtype>
inline int operator != (const tCubeBase<coordtype> &c1,
	const tCubeBase<coordtype> &c2)
{
	return c1. n != c2. n;
} /* operator != */

// --------------------------------------------------

// Compute the Cartesian product of two cubes.
template <class coordtype>
inline tCubeBase<coordtype> operator * (const tCubeBase<coordtype> &c1,
	const tCubeBase<coordtype> &c2)
//End/
{
	int dim1 = c1. dim (), dim2 = c2. dim ();
	if (dim1 + dim2 >= tCubeBase<coordtype>::MaxDim)
		throw "Dimension too high to concatenate coordinates.";
	coordtype coord [tCubeBase<coordtype>::MaxDim];
	return tCubeBase<coordtype> (coordConcat (coord, c1. coord (), dim1,
		c2. coord (), dim2), dim1 + dim2);
} /* operator * */

} // namespace homology
} // namespace capd


#endif // _CUBEBASE_H_

/// @}
