/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cubefix.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: May 17, 2006.


#ifndef _CAPD_CUBEFIX_H_
#define _CAPD_CUBEFIX_H_

#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/pointset.h"
#include "homology/chains.h"
#include "homology/bitfield.h"
#include "homology/integer.h"
#include "homology/hashsets.h"
#include "homology/gcomplex.h"
#include "homology/pointbas.h"
#include "homology/cubemain.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstring>

namespace capd {
namespace homology {

// a friend class template
template <int dimfix, class coordtype>
class tCellFix;

// --------------------------------------------------
// ------------ Cube of fixed dimension -------------
// --------------------------------------------------
// An n-dimensional cube of size 1 in R^n
// with integer coordinates and edges parallel to the axes.

template <int dimfix, class coordtype>
class tCubeFix
{
public:
	// the type of the coordinates
	typedef coordtype CoordType;

	// the maximal dimension of a cube (unused)
	static const int MaxDim = 512;

	// default constructor
	tCubeFix ();
	
	// constructor from a table of coordinates
	tCubeFix (const coordtype *c, int dim = 0);

	// constructor from a number (invalid in this class)
	tCubeFix (int number, int dim);

	// copy constructor
	tCubeFix (const tCubeFix<dimfix,coordtype> &c);

	// assignment operator
	tCubeFix<dimfix,coordtype> &operator =
		(const tCubeFix<dimfix,coordtype> &c);

	// return the dimension of the cube
	int dim () const;

	// return the dimension of the cube
	operator int () const;

	// return the coordinates of the cube
	const coordtype *coord () const;

	// fill out the coordinate table with the cube's coordinates
	template <class intType>
	intType *coord (intType *c) const;

	// return the coordinates of the cube
	operator const coordtype * () const;

	// hashing key no. 1 required by the hashing set template
	unsigned hashkey1 () const;

	// hashing key no. 2 required by the hashing set template
	unsigned hashkey2 () const;

	// the name of the objects represented by this class
	static const char *name ();

	// the plural name of the objects represented by this class
	static const char *pluralname ();

	// friend(s)
	friend class tCellFix<dimfix, coordtype>;

private:
	// a table with the coordinates of the minimal vertex of the cube
	coordtype tab [dimfix];

}; /* class tCubeFix */

// --------------------------------------------------

template <int dimfix, class coordtype>
inline tCubeFix<dimfix,coordtype>::tCubeFix ()
{
	return;
} /* tCubeFix::tCubeFix */

template <int dimfix, class coordtype>
inline tCubeFix<dimfix,coordtype>::tCubeFix
	(const coordtype *c, int dim)
{
	if (dim && ((dim != dimfix) || (dim < 0)))
		throw "Wrong dimension of a fixed-dim cube.";
	tPointBase<coordtype>::wrapcopy (tab, c, dimfix);
	tPointBase<coordtype>::outOfControl = true;
	return;
} /* tCubeFix::tCubeFix */

template <int dimfix, class coordtype>
inline tCubeFix<dimfix,coordtype>::tCubeFix (int, int)
{
	throw "Unable to construct a cube from a number.";
} /* tCubeFix::tCubeFix */

template <int dimfix, class coordtype>
inline tCubeFix<dimfix,coordtype>::tCubeFix
	(const tCubeFix<dimfix,coordtype> &c)
{
	for (int i = 0; i < dimfix; ++ i)
		tab [i] = c. tab [i];
	return;
} /* tCubeFix::tCubeFix */

template <int dimfix, class coordtype>
inline tCubeFix<dimfix,coordtype> &tCubeFix<dimfix,coordtype>::operator =
	(const tCubeFix<dimfix,coordtype> &c)
{
	for (int i = 0; i < dimfix; ++ i)
		tab [i] = c. tab [i];
	return *this;
} /* tCubeFix::operator = */

template <int dimfix, class coordtype>
inline int tCubeFix<dimfix,coordtype>::dim () const
{
	return dimfix;
} /* tCubeFix::dim */

template <int dimfix, class coordtype>
inline tCubeFix<dimfix,coordtype>::operator int () const
{
	return dimfix;
} /* tCubeFix::operator int */

template <int dimfix, class coordtype>
inline const coordtype *tCubeFix<dimfix,coordtype>::coord () const
{
	return tab;
} /* tCubeFix::coord */

template <int dimfix, class coordtype>
template <class intType>
inline intType *tCubeFix<dimfix,coordtype>::coord (intType *c) const
{
	for (int i = 0; i < dimfix; ++ i)
		c [i] = static_cast<intType> (tab [i]);
	return c;
} /* tCubeFix::coord */

template <int dimfix, class coordtype>
inline tCubeFix<dimfix,coordtype>::operator const coordtype * () const
{
	return tab;
} /* tCubeFix::operator coordtype * */

template <int dimfix, class coordtype>
inline unsigned tCubeFix<dimfix,coordtype>::hashkey1 () const
{
	switch (dimfix)
	{
	case 1:
		return (unsigned) (tab [0]) << 12;
	case 2:
		return (((unsigned) (tab [0])) << 18) +
			(((unsigned) (tab [1])) << 6);
	default:
		return (((unsigned) (tab [0])) << 18) +
			(((unsigned) (tab [1])) << 6) +
			(((unsigned) (tab [2])) >> 6);
	}
} /* tCubeFix::hashkey1 */

template <int dimfix, class coordtype>
inline unsigned tCubeFix<dimfix,coordtype>::hashkey2 () const
{
	switch (dimfix)
	{
	case 1:
		return (unsigned) (tab [0]) << 3;
	case 2:
		return ((unsigned) (tab [0]) >> 1) +
			((unsigned) (tab [1]) << 13);
	default:
		return (((unsigned) (tab [dimfix - 1])) << 20) +
			(((unsigned) (tab [dimfix - 2])) << 9) +
			(((unsigned) (tab [dimfix - 3])) >> 1);
	}
} /* tCubeFix::hashkey2 */

template <int dimfix, class coordtype>
const char *tCubeFix<dimfix,coordtype>::name ()
{
	return "cube";
} /* tCubeFix::name */

template <int dimfix, class coordtype>
const char *tCubeFix<dimfix,coordtype>::pluralname ()
{
	return "cubes";
} /* tCubeFix::pluralname */

template <int dim1, int dim2, class coordtype>
inline int operator == (const tCubeFix<dim1,coordtype> &c1,
	const tCubeFix<dim2,coordtype> &c2)
{
	return (dim1 == dim2) && thesame (c1. coord (), c2. coord (), dim1);
} /* operator == */

template <int dim1, int dim2, class coordtype>
inline int operator != (const tCubeFix<dim1,coordtype> &c1,
	const tCubeFix<dim2,coordtype> &c2)
{
	return !(c1 == c2);
} /* operator != */

// --------------------------------------------------

// Compute the Cartesian product of two cubes.
template <int dim1, int dim2, class coordtype>
inline tCubeFix<dim1+dim2,coordtype> operator *
	(const tCubeFix<dim1,coordtype> &c1,
	const tCubeFix<dim2,coordtype> &c2)
//End/
{
	coordtype coord [dim1 + dim2];
	return tCubeFix<dim1+dim2,coordtype>
		(coordConcat<dim1,dim2,coordtype> (coord,
		c1. coord (), c2. coord ()));
} /* operator * */

} // namespace homology
} // namespace capd


#endif // _CUBEFIX_H_

/// @}
