/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cubemain.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: November 5, 2004.


#ifndef _CAPD_CUBEMAIN_H_
#define _CAPD_CUBEMAIN_H_


#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/pointset.h"
#include "homology/chains.h"
#include "homology/bitfield.h"
#include "homology/integer.h"
#include "homology/hashsets.h"
#include "homology/gcomplex.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstring>

namespace capd {
namespace homology {

// Concatenate the coordinates into one big table.
template <class coordtype>
inline coordtype *coordConcat (coordtype *coord,
	const coordtype *coord1, int dim1, const coordtype *coord2, int dim2)
{
	int i, j = 0;
	for (i = 0; i < dim1; i ++)
		coord [j ++] = coord1 [i];
	for (i = 0; i < dim2; i ++)
		coord [j ++] = coord2 [i];
	return coord;
} /* coordConcat */

// Concatenate the coordinates into one big table.
template <int dim1, int dim2, class coordtype>
inline coordtype *coordConcat (coordtype *coord,
	const coordtype *coord1, const coordtype *coord2)
{
	int i, j = 0;
	for (i = 0; i < dim1; i ++)
		coord [j ++] = coord1 [i];
	for (i = 0; i < dim2; i ++)
		coord [j ++] = coord2 [i];
	return coord;
} /* coordConcat */

// --------------------------------------------------

// a template for converting between various cube types
template <class dest_cube, class src_cube>
inline dest_cube cube_cast (const src_cube &src)
{
	const typename src_cube::CoordType *coord = src. coord ();
	if (coord)
		return dest_cube (coord, src. dim ());
	typename dest_cube::CoordType tab [src_cube::MaxDim];
	return dest_cube (src. coord (tab), src. dim ());
} /* cube_cast */

// a template for converting between various cube types
// to be used if the types of coordinates differ
template <class dest_cube, class src_cube>
inline dest_cube cube_castcoord (const src_cube &src)
{
	typename dest_cube::CoordType tab [src_cube::MaxDim];
	return dest_cube (src. coord (tab), src. dim ());
} /* cube_castcoord */

// --------------------------------------------------

// Write the cube to the output stream in the text mode.
template <class cubetype>
inline std::ostream &WriteCube (std::ostream &out, const cubetype &c)
{
	const typename cubetype::CoordType *coord = c. coord ();
	int dim = c. dim ();
	out << "(";
	while (dim --)
	{
		out << *(coord ++);
		if (dim)
			out << ",";
	}
	out << ")";
	return out;
} /* WriteCube */

// --------------------------------------------------

// Read a cube from the input text stream.
template <class cubetype>
std::istream &ReadCubeFix (std::istream &in, cubetype &c, int dimfix)
{
	// retrieve the type of coordinates of the cube to read
	typedef typename cubetype::CoordType coordtype;

	// ignore any comments, spaces, tabs and new-line characters
	ignorecomments (in);

	// if there is a number in the input, then there are apparently
	// no parentheses used for the coords or the cube is defined
	// by its number (also indicated with '#')
	if (((in. peek () >= '0') && (in. peek () <= '9')) ||
		(in. peek () == '-') || (in. peek () == '+') ||
		(in. peek () == '#'))
	{
		bool cubenumber = false;
		if (in. peek () == '#')
		{
			in. get ();
			ignorecomments (in);
			cubenumber = true;
		}
		if (in. peek () == '+')
			in. get ();
		int n = -1;
		in >> n;
		while ((in. peek () == ' ') || (in. peek () == '\t'))
			in. get ();

		// if the next coordinate of the cube follows,
		// read the coordinates until end-of-line is encountered
		if (!cubenumber &&
			(((in. peek () >= '0') && (in. peek () <= '9')) ||
			(in. peek () == '-') || (in. peek () == '+')))
		{
			// read the coords and determine the space dimension
			coordtype coord [cubetype::MaxDim];
			coord [0] = n;
			int dim = -1;
			if (in. peek () != '\n')
			{
				dim = readcoordinates (in, coord + 1,
					cubetype::MaxDim - 1, '\n');
				if (dim < 0)
					throw "Unable to read a cube: "
						"Dim too high.";
			}
			++ dim;

			// if could not read the cube, throw an error message
			if ((dimfix >= 0) && (dim != dimfix))
				throw "Unable to read a cube: Wrong dim.";

			// create a cube with the given coordinates
			c = cubetype (coord, dim);

			return in;
		}

		// if the cube is given by its number, read it this way
		else
		{
			int dim = tPointBase<coordtype>::defaultdimension ();
			const coordtype *coord = 0;
			if ((n > 0) && (dim > 0))
				coord = tPointBase<coordtype>::coord (n - 1,
					dim);
			if (!coord)
				throw "Cube no. out of range while reading.";
			c = cubetype (n - 1, dim);
			return in;
		}
	}

	// read the coordinates and determine the space dimension
	coordtype coord [cubetype::MaxDim];
	int dim = readcoordinates (in, coord, cubetype::MaxDim);
	if (dim < 0)
		throw "Unable to read a cube: Dimension too high.";

	// if could not read the cube, throw an error message
	if ((dimfix >= 0) && (dim != dimfix))
		throw "Unable to read a cube: Wrong dimension.";

	// create the cube with the given coordinates
	c = cubetype (coord, dim);

	return in;
} /* ReadCubeFix */

template <class cubetype>
inline std::istream &ReadCube (std::istream &in, cubetype &c)
{
	return ReadCubeFix (in, c, -1);
} /* ReadCube */


} // namespace homology
} // namespace capd


#endif // _CUBEMAIN_H_

/// @}
