/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cubevar.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: May 17, 2006.


#ifndef _CAPD_CUBEVAR_H_
#define _CAPD_CUBEVAR_H_

#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/pointset.h"
#include "homology/chains.h"
#include "homology/bitfield.h"
#include "homology/integer.h"
#include "homology/hashsets.h"
#include "homology/gcomplex.h"
#include "homology/pointbas.h"
#include "homology/cubemain.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstring>

namespace capd {
namespace homology {

// a friend class template
template <class coordtype>
class tCellVar;

// --------------------------------------------------
// ----------- Cube with allocated memory -----------
// --------------------------------------------------
// An n-dimensional cube of size 1 in R^n
// with integer coordinates and edges parallel to the axes.

template <class coordtype>
class tCubeVar
{
public:
	// the type of the coordinates
	typedef coordtype CoordType;

	// the maximal dimension of a cube (used only in operator *)
	static const int MaxDim = 512;

	// default constructors
	tCubeVar ();
	
	// constructor from a table of coordinates
	tCubeVar (const coordtype *c, int dim);

	// constructor from a number (invalid in this class)
	tCubeVar (int number, int dim);

	// copy constructor
	tCubeVar (const tCubeVar<coordtype> &c);

	// assignment operator
	tCubeVar<coordtype> &operator = (const tCubeVar<coordtype> &c);

	// destructor
	~tCubeVar ();

	// return the dimension of the cube
	int dim () const;

	// return the dimension of the cube
	operator int () const;

	// return the coordinates of the cube
	const coordtype *coord () const;

	// fill out the coordinate table with the cube's coordinates
	template <class intType>
	intType *coord (intType *c) const;

	// return the coordinates of the cube
	operator const coordtype * () const;

	// hashing key no. 1 required by the hashing set template
	unsigned hashkey1 () const;

	// hashing key no. 2 required by the hashing set template
	unsigned hashkey2 () const;

	// the name of the objects represented by this class
	static const char *name ();

	// the plural name of the objects represented by this class
	static const char *pluralname ();

	// friend(s)
	friend class tCellVar<coordtype>;

private:
	// the table containing the coordinates of the cube,
	// as well as its dimension (at tab [0])
	coordtype *tab;

	// initialize the data for a cube of a given dimension > 0;
	// return the pointer to the coordinates table (i.e. tab + 1)
	coordtype *initialize (int dim);

}; /* class tCubeVar */

// --------------------------------------------------

template <class coordtype>
inline coordtype *tCubeVar<coordtype>::initialize (int d)
{
	tab = new coordtype [d + 1];
	if (!tab)
		throw "Not enough memory for a cube.";
	*tab = d;
	return (tab + 1);
} /* tCubeVar::initialize */

// --------------------------------------------------

template <class coordtype>
inline tCubeVar<coordtype>::tCubeVar ()
{
	tab = NULL;
	return;
} /* tCubeVar::tCubeVar */

template <class coordtype>
inline tCubeVar<coordtype>::tCubeVar (const coordtype *c, int d)
{
	if (d < 0)
		throw "Negative dimension of a cube.";
	if (d)
		tPointBase<coordtype>::wrapcopy (initialize (d), c, d);
	else
		tab = NULL;
	tPointBase<coordtype>::outOfControl = true;
	return;
} /* tCubeVar::tCubeVar */

template <class coordtype>
inline tCubeVar<coordtype>::tCubeVar (int, int)
{
	throw "Unable to construct a cube from a number.";
} /* tCubeVar::tCubeVar */

template <class coordtype>
inline tCubeVar<coordtype>::tCubeVar (const tCubeVar<coordtype> &c)
{
	if (!c. dim ())
		tab = NULL;
	else
		copycoord (initialize (c. dim ()), c. coord (), c. dim ());
	return;
} /* tCubeVar::tCubeVar */

template <class coordtype>
inline tCubeVar<coordtype> &tCubeVar<coordtype>::operator =
	(const tCubeVar<coordtype> &c)
{
	if (dim () == c. dim ())
		copycoord (tab + 1, c. coord (), dim ());
	else
	{
		if (tab)
			delete [] tab;
		if (c. dim ())
			copycoord (initialize (c. dim ()), c. coord (),
				c. dim ());
		else
			tab = NULL;
	}
	return *this;
} /* tCubeVar::operator = */

template <class coordtype>
inline tCubeVar<coordtype>::~tCubeVar<coordtype> ()
{
	if (tab)
		delete [] tab;
	return;
} /* tCubeVar::~tCubeVar */

template <class coordtype>
inline int tCubeVar<coordtype>::dim () const
{
	return tab ? *tab : 0;
} /* tCubeVar::dim */

template <class coordtype>
inline tCubeVar<coordtype>::operator int () const
{
	return dim ();
} /* operator int */

template <class coordtype>
inline const coordtype *tCubeVar<coordtype>::coord () const
{
	return tab ? (tab + 1) : NULL;
} /* tCubeVar::coord */

template <class coordtype>
template <class intType>
inline intType *tCubeVar<coordtype>::coord (intType *c) const
{
	if (!tab)
		return NULL;
	for (int i = 0; i < *tab; ++ i)
		c [i] = static_cast<intType> (tab [i + 1]);
	return c;
} /* tCubeVar::coord */

template <class coordtype>
inline tCubeVar<coordtype>::operator const coordtype * () const
{
	return coord ();
} /* tCubeVar::operator coordtype * */

template <class coordtype>
inline unsigned tCubeVar<coordtype>::hashkey1 () const
{
	int d = dim ();
	switch (d)
	{
	case 0:
		return 0;
	case 1:
		return (unsigned) (tab [0]) << 12;
	case 2:
		return (((unsigned) (tab [0])) << 18) +
			(((unsigned) (tab [1])) << 6);
	default:
		return (((unsigned) (tab [0])) << 18) +
			(((unsigned) (tab [1])) << 6) +
			(((unsigned) (tab [2])) >> 6);
	}
} /* tCubeVar::hashkey1 */

template <class coordtype>
inline unsigned tCubeVar<coordtype>::hashkey2 () const
{
	int d = dim ();
	switch (d)
	{
	case 0:
		return 1;
	case 1:
		return (unsigned) (tab [0]) << 3;
	case 2:
		return ((unsigned) (tab [0]) >> 1) +
			((unsigned) (tab [1]) << 13);
	default:
		return (((unsigned) (tab [d - 1])) << 20) +
			(((unsigned) (tab [d - 2])) << 9) +
			(((unsigned) (tab [d - 3])) >> 1);
	}
} /* tCubeVar::hashkey2 */

template <class coordtype>
inline const char *tCubeVar<coordtype>::name ()
{
	return "cube";
} /* tCubeVar::name */

template <class coordtype>
inline const char *tCubeVar<coordtype>::pluralname ()
{
	return "cubes";
} /* tCubeVar::pluralname */

// --------------------------------------------------

template <class coordtype>
inline int operator == (const tCubeVar<coordtype> &c1,
	const tCubeVar<coordtype> &c2)
{
	return (c1. dim () == c2. dim ()) &&
		thesame (c1. coord (), c2. coord (), c1. dim ());
} /* operator == */

template <class coordtype>
inline int operator != (const tCubeVar<coordtype> &c1,
	const tCubeVar<coordtype> &c2)
{
	return !(c1 == c2);
} /* operator != */

// --------------------------------------------------

// Compute the Cartesian product of two cubes.
template <class coordtype>
inline tCubeVar<coordtype> operator * (const tCubeVar<coordtype> &c1,
	const tCubeVar<coordtype> &c2)
//End/
{
	int dim1 = c1. dim (), dim2 = c2. dim ();
	if (dim1 + dim2 >= tCubeVar<coordtype>::MaxDim)
		throw "Dimension too high to concatenate coordinates.";
	coordtype coord [tCubeVar<coordtype>::MaxDim];
	return tCubeVar<coordtype> (coordConcat (coord,
		c1. coord (), dim1, c2. coord (), dim2), dim1 + dim2);
} /* operator * */

} // namespace homology
} // namespace capd


#endif // _CUBEVAR_H_

/// @}
