/// @addtogroup matrixAlgorithms
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file floatMatrixAlgorithms.hpp
///
/// @author The CAPD Group
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _floatMatrixAlgorithms_hpp_
#define _floatMatrixAlgorithms_hpp_

#include <vector>
#include <stdexcept>
#include <iostream>
#include "capd/minmax.h"
#include "capd/power.h"

template<typename IntervalType>
inline IntervalType Mid(const IntervalType& x)
{
   return x.mid();
}

template<>
inline double Mid(const double& x){return x;}

template<>
inline long Mid(const long& x){return x;}

template<>
inline int Mid(const int& x){return x;}

// ------------- Gauss elimination ------------------------------- //

template<typename MatrixType>
typename MatrixType::RowVectorType gauss(MatrixType a, typename MatrixType::RowVectorType b)
{
// on output result = a^-1 b
   if(a.numberOfRows()!=a.numberOfColumns() || a.numberOfRows()!=b.dimension())
      throw std::runtime_error("Call gauss elimination for nonsquare matrix");

   typedef typename MatrixType::RowVectorType VectorType;
   typedef typename MatrixType::ScalarType ScalarType;

   VectorType result(a.numberOfRows());
   int i,j,k;
   int dimension = b.dimension();
   std::vector<int> p (dimension);
   for(i=0;i<dimension;++i) p[i]=i;

   for(j=0;j<dimension-1;++j)
   {
      ScalarType a_max,a_max_temp;
      int i_max,temp;
      a_max = capd::abs(a[p[j]][j]);
      i_max = j;
      temp = p[j];
      for(i=j+1;i<dimension;++i)
      {
         if( (a_max_temp = capd::abs(a[p[i]][j])) > a_max)
         {
            a_max = a_max_temp;
            i_max = i;
         }
      }
      p[j] = p[i_max];
      p[i_max] = temp;

      ScalarType divisor = a[p[j]][j];
      if( isSingular(divisor) )
      {
         throw std::runtime_error( "Gauss elimination: singular matrix\n");
      }
      for(i=j+1;i<dimension;++i)
      {
         ScalarType factor = a[p[i]][j]/a[p[j]][j];
         a[p[i]][j] = static_cast<ScalarType>(0);
         b[p[i]] -= factor*b[p[j]];
         for(k=j+1;k<dimension;++k)
         {
            a[p[i]][k] -= factor*a[p[j]][k];
         }
      }
   }

   if( isSingular(a[p[dimension-1]][dimension-1]) )
   {
      throw std::runtime_error( "Gauss elimination: singular matrix\n");
   }
   for(i=dimension-1;i>=0;--i)
   {
      result[i] = b[p[i]];
      for(j=i+1;j<dimension;++j)
         result[i] -= a[p[i]][j]*result[j];
      result[i] /= a[p[i]][i];
   }
   return result;
}


// Gramm-Schmit column orthonormalization
template<typename MatrixType>
void orthonormalize(MatrixType& Q)
{
   typedef typename MatrixType::RowVectorType VectorType;
   typedef typename MatrixType::ScalarType ScalarType;
   typedef typename VectorType::intVectorType intVectorType;
   typedef typename MatrixType::RefColumnVectorType ColumnVectorType;

   VectorType norm(Q.numberOfColumns(),true);
   intVectorType p(Q.numberOfColumns(),true);

   int i=0;
   typename MatrixType::ContainerIterator e=Q.cend();
   typename VectorType::iterator bn = norm.begin(), en=norm.end();
   while(bn!=en)
   {
      ScalarType s(0);
      typename MatrixType::ContainerIterator b=Q.cbegin()+i;
      while(b!=e)
      {
         s += power(*b,2);
         b += Q.rowStride();
      }
      *bn = s;
      ++bn;
      ++e;
      ++i;
   }

   norm.sorting_permutation(p);
   typename intVectorType::iterator bp = p.begin(), ep = p.end();
   while(bp!=ep)
   {
      ColumnVectorType qi = Q.column(*bp);
      if(qi.normalize())
      {
         typename intVectorType::iterator bj = ++bp;
         while(bj!=ep)
         {
            ColumnVectorType qj = Q.column(*bj);
            qj -= (qj * qi) * qi;
            ++bj;
         }
      }else{
         throw std::runtime_error("Failed to decompose matrix");
      }
   }
}

template<typename MatrixType>
void QR_decompose(const MatrixType& A, MatrixType& Q, MatrixType& R)
{
   Q = A;
   int i,j, dimension = A.numberOfColumns();
   for(i=0;i<dimension;++i)
   {
      if(Q.column(i).normalize())
      {
         for(j=i+1;j<dimension;++j)
         {
            Q.column(j) -= (Q.column(j)*Q.column(i)) * Q.column(i);
         }
      }else{
         throw std::runtime_error("Failed to decompose A");
      }
   }
   R = Transpose(Q) * A;
}

template<typename MatrixType>
int diagonalize(const MatrixType& A, MatrixType& D)
{
   typedef typename MatrixType::RowVectorType VectorType;
   typedef typename MatrixType::ScalarType ScalarType;
   int dimension = A.numberOfRows();
   static ScalarType diagonalizingRelTolerance(0.00001);
   MatrixType lastD(dimension,dimension);

   ScalarType last_sum_of_squares(10000000000000000.0);
   D=A;
   if(dimension == 1) return 1;
   ScalarType s(0), d(0);
   int i,j,p=0,q=0;
   for(i=1;i<dimension;++i)
      for(j=0;j<i;++j)
         s += power(D[i][j],2);
   for(i=0;i<dimension;++i)
      d += power(D[i][i],2);
   if( !(d>0) ) return 1;

   while( !(s<d*diagonalizingRelTolerance) )
   {
      if(!(right(s)<right(last_sum_of_squares)))
      {
         D = lastD;
         break;
      }
      ScalarType dominant(0);
      for(i=1;i<dimension;++i)
         for(j=0;j<i;++j)
         {
            ScalarType element = left(capd::abs(D[i][j]));
            if( element > dominant )
            {
               p = i;
               q = j;
               dominant = element;
            }
         }
      if( !(dominant!= ScalarType(0)) ) break;  //procceeding with iteration can only worsen things

      ScalarType alpha = (Mid(D[q][q]) - Mid(D[p][p]))/(ScalarType(2)*Mid(D[p][q]));
      ScalarType beta = (alpha>0 ? alpha + sqrt(power(alpha,2)+1) : alpha - sqrt(power(alpha,2)+1));
      ScalarType sinth = (alpha>0 ? ScalarType(1)/sqrt(power(beta,2)+ScalarType(1))
                                  : ScalarType(-1)/sqrt(power(beta,2)+ScalarType(1))
         );
      ScalarType costh = beta*sinth;
      VectorType ap = D[p];
      VectorType aq = D[q];
      lastD = D;
      D[p][p] = aq[q]*power(sinth,2) - ScalarType(2)*ap[q]*sinth*costh + ap[p]*power(costh,2);
      D[q][q] = aq[q]*power(costh,2) + ScalarType(2)*ap[q]*sinth*costh + ap[p]*power(sinth,2);
      D[p][q] = D[q][p] = ap[q]*(power(costh,2) - power(sinth,2)) + (ap[p]-aq[q])*sinth*costh;
      for(j=0;j<dimension;++j)
         if(j!=p && j!=q)
            D[p][j] = D[j][p] = -aq[j]*sinth+ap[j]*costh;
      for(j=0;j<dimension;++j)
         if(j!=p && j!=q)
            D[q][j] = D[j][q] = aq[j]*costh+ap[j]*sinth;

// last_sum_of_squares=s+d;
      last_sum_of_squares = s;
      d = ScalarType(0);
      for(i=0;i<dimension;++i) d += power(D[i][i],2);
         if( !(d!=0) ) return 0;

      s=ScalarType(0);
      for(i=1;i<dimension;++i)
         for(j=0;j<i;++j)
          s += power(D[i][j],2);

      if( !(s!=0) ) break;  //procceeding with iteration can only worsen things
   }
   return 1;
}

template<typename MatrixType>
typename MatrixType::ScalarType spectralRadius(const MatrixType &A)
{
   typedef typename MatrixType::ScalarType ScalarType;
   static ScalarType diagonalizingRelTolerance(0.00001);
   int dimension = A.numberOfRows();
   MatrixType D(dimension,dimension);
   if(!diagonalize(A,D))
   {
      std::runtime_error("spectralRadius::Failed to diagonalize a matrix");
   }
   ScalarType dominant(0), d(0);
   for(int i=0;i<dimension;++i)
   {
      ScalarType elem = capd::abs(D[i][i]);
      if(!(elem<dominant)) dominant = elem;
      d += power(D[i][i],2);
   }
   return dominant + d*diagonalizingRelTolerance;
}


template<typename MatrixType>
typename MatrixType::ScalarType maxEigenValue(const MatrixType &A)
{
   typedef typename MatrixType::ScalarType ScalarType;
   int dimension = A.numberOfRows();
   MatrixType D(dimension,dimension);
   if(!diagonalize(A,D))
   {
      throw std::runtime_error("maxEigenValue::Failed to diagonalize a matrix");
   }
   ScalarType dominant(0);
   for(int i=0;i<dimension;++i)
   {
      ScalarType diam(0);
      for(int j=0;j<dimension;++j)
         if( i != j )
            diam += capd::abs(D[i][j]);
      ScalarType elem = right(D[i][i]+diam);
      if(i==0) dominant=elem;
      else if(!(elem<dominant)) dominant = elem;
   }
   return dominant;
}

/**
  Crout Decomposition of a matrix
  As a result matrix D is a lower triangle
  and G is an upper triangle with 1 on diagonal
*/
template<typename MatrixType>
void croutDecomposition(const MatrixType &A, MatrixType &D, MatrixType &G)
{
   D.clear();
   G.clear();
   int i,j,k;
   for(j=0;j<D.numberOfRows();j++)
   {
      for(i=j;i<D.numberOfColumns();i++)
      {
         D[i][j] = A[i][j];
         for(k=0;k<=j-1;k++)
            D[i][j] -= D[i][k]*G[k][j];
      }
      G[j][j]=1;
      for(i=j+1;i<G.numberOfRows();i++)
      {
         G[j][i] = A[j][i];
         for(k=0;k<=j-1;k++)
            G[j][i] -= D[j][k]*G[k][i];
         if(!(D[j][j]!=0))
         {
            throw std::runtime_error("Crout Decomposition: cannot decompose singular matrix");
         }
         G[j][i] /= D[j][j];
      }
   }
}

template<typename MatrixType>
MatrixType invLowerTriangleMatrix(const MatrixType &A)
{
   MatrixType result(A.numberOfRows(),A.numberOfColumns());
   typedef typename MatrixType::ScalarType Scalar;
   for(int i=0;i<result.numberOfRows();++i)
   {
      if(!(A[i][i]<0 || A[i][i]>0))
      {
         throw std::runtime_error("cannot inverse lower triangle matrix - zero at diagonal");
      }
      result[i][i] = Scalar(1) / A[i][i];
      for(int j=0;j<i;++j)
      {
         for(int k=j;k<i;++k)
            result[i][j] += A[i][k] * result[k][j];
         result[i][j] *= -result[i][i];
      }
   }
   return result;
}

template<typename MatrixType>
MatrixType invUpperTriangleMatrix(const MatrixType &A)
{
   int dim = A.numberOfRows();
   MatrixType result(dim,A.numberOfColumns());
   typedef typename MatrixType::ScalarType Scalar;
   for(int i=0;i<result.numberOfRows();++i)
   {
      if(!(A[i][i]<0 || A[i][i]>0))
      {
         throw std::runtime_error("cannot inverse upper triangle matrix - zero at diagonal");
      }
      result[i][i] = Scalar(1) / A[i][i];
      for(int j=0;j<i;++j)
      {
         for(int k=j;k<i;++k)
            result[j][i] += A[k][i] * result[j][k];
         result[j][i] *= -result[i][i];
      }
   }
   return result;
}


template<typename MatrixType>
MatrixType inverseMatrix(const MatrixType &A)
{
   if(A.numberOfRows()!=A.numberOfColumns())
   {
      throw std::runtime_error("Cannot inverse nonsquare matrix!");
   }
   MatrixType Q(A.numberOfRows(),A.numberOfColumns()),
              R(A.numberOfRows(),A.numberOfColumns());
   QR_decompose(A,Q,R);
   R = invUpperTriangleMatrix(R);
   Q.Transpose();
   return R*Q;
}

template<typename MatrixType>
MatrixType gaussInverseMatrix(const MatrixType A)
{
   if(A.numberOfRows()!=A.numberOfColumns())
   {
      throw std::runtime_error("Cannot inverse nonsquare matrix!");
   }
   int dim = A.numberOfRows();
   MatrixType A_1(dim, dim);
   typename MatrixType::RowVectorType b(dim);

   for(int i=0; i<dim; i++)
   {
      for(int j=0; j<dim; j++)
         b[j]=(typename MatrixType::ScalarType)(j==i);
      A_1[i]=gauss(A, b);
   }
   A_1.Transpose();
   return A_1;
}

#endif //_floatMatrixAlgorithms_hpp_

/// @}
