/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file columnVector.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _ColumnVector_H
#define _ColumnVector_H

#include <ostream>

namespace capd{
namespace vectalg{

template<typename Scalar>
class ColumnIterator{
public:
   typedef Scalar ScalarType;

   inline ColumnIterator(ScalarType *a_p, int a_pointerStride)
      : m_Pointer(a_p), m_pointerStride(a_pointerStride)
   {}

   inline ColumnIterator operator++(int)
   {
      ColumnIterator it(*this);
      m_Pointer += m_pointerStride;
      return it;
   }

   inline ColumnIterator operator--(int)
   {
      ColumnIterator it(*this);
      m_Pointer -= m_pointerStride;
      return it;
   }

   inline ColumnIterator& operator++()
   {
      m_Pointer += m_pointerStride;
      return *this;
   }

   inline ColumnIterator& operator--()
   {
      m_Pointer -= m_pointerStride;
      return *this;
   }

   inline ColumnIterator& operator+=(int jump)
   {
      m_Pointer+= jump*m_pointerStride;
      return *this;
   }

   inline ColumnIterator& operator-=(int jump)
   {
      m_Pointer-= jump*m_pointerStride;
      return *this;
   }

   inline bool operator!=(const ColumnIterator& second)
   {
      return m_Pointer!=second.m_Pointer;
   }

   inline ScalarType& operator*()
   {
      return *m_Pointer;
   }

   inline ScalarType* operator->()
   {
      return m_Pointer;
   }

   private:
   ScalarType *m_Pointer;
   int m_pointerStride;
   ColumnIterator(){} // we do not need a default constructor
};

// --------------------- const_iterator -------------------

template<typename Scalar>
class const_ColumnIterator{
public:
   typedef Scalar ScalarType;

   inline const_ColumnIterator(const ScalarType* a_p, int a_pointerStride)
      : m_Pointer(a_p), m_pointerStride(a_pointerStride)
   {}

   inline const_ColumnIterator operator++(int)
   {
      const_ColumnIterator it(*this);
      m_Pointer += m_pointerStride;
      return it;
   }

   inline const_ColumnIterator operator--(int)
   {
      const_ColumnIterator it(*this);
      m_Pointer -= m_pointerStride;
      return it;
   }

   inline const_ColumnIterator& operator++()
   {
      m_Pointer += m_pointerStride;
      return *this;
   }

   inline const_ColumnIterator& operator--()
   {
      m_Pointer -= m_pointerStride;
      return *this;
   }

   inline const_ColumnIterator& operator+=(int jump)
   {
      m_Pointer += jump*m_pointerStride;
      return *this;
   }

   inline const_ColumnIterator& operator-=(int jump)
   {
      m_Pointer -= jump*m_pointerStride;
      return *this;
   }

   inline bool operator!=(const const_ColumnIterator& second)
   {
      return m_Pointer!=second.m_Pointer;
   }

   inline const ScalarType& operator*()
   {
      return *m_Pointer;
   }

   inline const ScalarType* operator->()
   {
      return m_Pointer;
   }
   private:
   const ScalarType*  m_Pointer;
   const_ColumnIterator(){} // we do not need a default constructor
   int m_pointerStride;
};

}} // namespace capd::vectalg


template<typename Scalar, int rows, int cols>
class Matrix;

template<typename Scalar,int dim>
class Vector;

template<typename Scalar, int rows>
class ColumnVector
{
public:
   typedef Scalar ScalarType;
   typedef capd::vectalg::ColumnIterator<Scalar> iterator;
   typedef capd::vectalg::const_ColumnIterator<Scalar> const_iterator;
   typedef ColumnVector VectorType;

   inline ColumnVector(const Scalar* a_pointer, int a_stride, int a_dim)
      : m_pointer(const_cast<Scalar*>(a_pointer)),
        m_stride(a_stride), m_dim(a_dim)
   {}
   ColumnVector& operator=(const ColumnVector&);
   ColumnVector& operator=(const Vector<Scalar,rows>&);
   ColumnVector& operator+=(const ColumnVector&);
   ColumnVector& operator+=(const Vector<Scalar,rows>&);
   ColumnVector& operator-=(const ColumnVector&);
   ColumnVector& operator-=(const Vector<Scalar,rows>&);
   operator Vector<Scalar,rows>() const;

   inline Scalar& operator[](int a_row)
   {
      return *(m_pointer + a_row*m_stride);
   }
   inline const Scalar& operator[](int a_row) const
   {
      return *(m_pointer + a_row*m_stride);
   }

   Scalar euclNorm() const;
   bool normalize();
   void clear();
   inline int dimension() const
   {
      return m_dim;
   }
   inline iterator begin()
   {
      return iterator(m_pointer,m_stride);
   }
   inline iterator end()
   {
      return iterator(m_pointer+m_dim*m_stride, m_stride);
   }
   inline const_iterator begin() const
   {
      return const_iterator(m_pointer, m_stride);
   }
   inline const_iterator end() const
   {
      return const_iterator(m_pointer+m_dim*m_stride, m_stride);
   }
protected:
   Scalar* m_pointer;
   int m_stride, m_dim;
}; // ----------------------------------------------------

template<typename Scalar, int rows>
Vector<Scalar,rows> operator+(const Vector<Scalar,rows>&, const ColumnVector<Scalar,rows>&);

template<typename Scalar, int rows>
Vector<Scalar,rows> operator+(const ColumnVector<Scalar,rows>&, const Vector<Scalar,rows>&);

template<typename Scalar, int rows>
Vector<Scalar,rows> operator+(const ColumnVector<Scalar,rows>&, const ColumnVector<Scalar,rows>&);


template<typename Scalar, int rows>
Vector<Scalar,rows> operator-(const Vector<Scalar,rows>&, const ColumnVector<Scalar,rows>&);

template<typename Scalar, int rows>
Vector<Scalar,rows> operator-(const ColumnVector<Scalar,rows>&, const Vector<Scalar,rows>&);

template<typename Scalar, int rows>
Vector<Scalar,rows> operator-(const ColumnVector<Scalar,rows>&, const ColumnVector<Scalar,rows>&);



template<typename Scalar, int rows>
Scalar operator*(const Vector<Scalar,rows>&, const ColumnVector<Scalar,rows>&);

template<typename Scalar, int rows>
Scalar operator*(const ColumnVector<Scalar,rows>&, const Vector<Scalar,rows>&);

template<typename Scalar, int rows>
Scalar operator*(const ColumnVector<Scalar,rows>&, const ColumnVector<Scalar,rows>&);



template<typename Scalar, int rows, int cols>
Vector<Scalar,rows> operator*(const Matrix<Scalar,rows,cols>&, const ColumnVector<Scalar,cols>&);

template<typename Scalar, int rows>
Vector<Scalar,rows> operator-(const ColumnVector<Scalar,rows>&);



template<typename Scalar, int rows>
Vector<Scalar,rows> operator*(const Scalar&, const ColumnVector<Scalar,rows>&);

template<typename Scalar, int rows>
Vector<Scalar,rows> operator*(const ColumnVector<Scalar,rows>&, const Scalar&);

template<typename Scalar, int rows>
Vector<Scalar,rows> operator/(const ColumnVector<Scalar,rows>&, const Scalar&);

template<typename Scalar, int rows>
std::ostream& operator << (std::ostream&, const ColumnVector<Scalar,rows>&);

#endif

/// @}
